import yaml
import re
import copy
import os
import math

from lxml import etree

def _recur_scad_config(fd, base, variables):

    if isinstance(variables, dict):
        for key, item in variables.items():
            if base == "":
                newbase = key
            else:
                newbase = base + "_" + key

            _recur_scad_config(fd, newbase, item)

            if base == "":
                fd.write("\n")

    elif isinstance(variables, (list, tuple)):
        fd.write(
            "{} = [{}];\n".format(base, ",".join(str(x) for x in variables)))

    elif isinstance(variables, (int, float)):
        fd.write("{} = {};\n".format(base, variables))

    else:
        raise Exception("unknown data", variables)


def generate_scad_config(filename, variables):
    print("generating scad file: " + filename)
    with open(filename, "w") as fd:
        fd.write("// rendered by script, do not edit manually!\n")
        _recur_scad_config(fd, "", variables)


def generate_scad_joint_pos(fd, key, joint):
    fd.write("module " + key + "_joint_pos()")
    fd.write(" translate({})".format(joint['pos']))
    fd.write(" rotate({})".format(joint['rot']))
    fd.write("{\n")
    module_args = joint['scad']['args']
    module_args = ",".join([str(i) for i in module_args])
    fd.write("  {}({});\n".format(joint['scad']['module'], module_args))
    fd.write("  children();")
    fd.write("\n}\n")
    fd.write("\n")


def generate_scad_leg_pos(filename, legs):
    with open(filename, "w") as fd:
        fd.write("include <leg_modules.scad>;\n")
        for leg_key, leg in legs.items():
            for joint_key, joint in leg['joints'].items():
                generate_scad_joint_pos(fd, leg_key + "_" + joint_key, joint)

            fd.write("module {}_leg()".format(leg_key))
            fd.write("translate({})".format(leg['pos']))
            fd.write("rotate({})".format(leg['rot']))
            fd.write("mirror({})".format(leg['mirror']))
            fd.write("{\n")

            for key in leg['joint_chain']:
                fd.write(" {}_{}_joint_pos()".format(leg_key, key))
            fd.write(";\n}\n")

        fd.write("module legs(){\n")
        for leg_key in legs.keys():
            fd.write(" {}_leg();\n".format(leg_key))
        fd.write("}\n")


        fd.write("module leg_pos(){\n")
        for leg_key, leg in legs.items():
            fd.write("translate({})".format(leg['pos']))
            fd.write("rotate({})".format(leg['rot']))
            fd.write("mirror({})".format(leg['mirror']))
            fd.write("children();\n")

        fd.write("}\n")

def check_stl(filename, module):

    assert filename[-1] == 'b'

    scad = "../scad/tmp.scad"

    with open(scad, "w") as fd:
        fd.write("include <model.scad>;\n")
        fd.write(module + "();")

    tmp_stl = "../stl/tmp.stl"

    os.system(" ".join(['openscad', '-o', tmp_stl, scad]))

    os.system("ivcon {} {}".format(tmp_stl, filename))

    os.remove(tmp_stl)
    os.remove(scad)


def generate_link(robot, name, scad_module):
    link = etree.SubElement(robot, 'link', name=name)

    visual = etree.SubElement(link, 'visual')

    geometry = etree.SubElement(visual, 'geometry')

    if scad_module is not None:
        stl_filename = 'stl/' + name + ".stlb"

        if not os.path.exists("../" + stl_filename):
            check_stl("../" + stl_filename, scad_module)

        etree.SubElement(
            geometry, 'mesh', filename="package://schpin_tote/" + stl_filename)

        material = etree.SubElement(
            visual, 'material', name=name + "_material")

        etree.SubElement(
            material, 'color', rgba="{} {} {} 1".format(*get_color(name)))

    return link


def get_color(name):
    m = hashlib.md5()
    m.update(name.encode('utf8'))
    h = m.digest()

    return [x / 256 for x in h[:3]]


def generate_urdf_joint(robot,
                        name,
                        parent,
                        child,
                        pos,
                        rot,
                        axis,
                        static=False):
    joint = etree.SubElement(robot, 'joint', name=name, type='revolute')

    etree.SubElement(joint, 'parent', link=parent.name)
    etree.SubElement(joint, 'child', link=child.name)

    if not static:
        etree.SubElement(
            joint,
            'limit',
            effort="1000",
            velocity="50",
            lower="-2",
            upper="2")

    etree.SubElement(
        joint,
        'origin',
        xyz="{} {} {}".format(* [x for x in pos]),
        rpy="{} {} {}".format(* [math.radians(x) for x in rot]))

    if axis is not None:
        etree.SubElement(joint, 'axis', xyz="{} {} {}".format(*axis))


" return number of replaced items, unreplaced items, new value "


def _parse_val(val, top_cfg):

    res = re.match("^\s*%(.*)%\s*$", val)
    if res is None:
        return 0, 0, val

    expr = res.group(1)
    base_expr = expr
    #print expr

    vars = re.findall("[a-zA-Z][a-zA-Z0-9\._]+", expr)

    replaced_count = 0
    unreplaced_count = 0

    for var in vars:
        if var in ["max", "min", "abs"]:
            continue
        keys = var.split('.')
        value = top_cfg
        while len(keys) > 0:
            if isinstance(value, list):
                value = value[int(keys[0])]
            elif isinstance(value, dict):
                try:
                    value = value[keys[0]]
                except KeyError:
                    raise Exception(
                        "Using key for variable that does not exist ", keys[0],
                        " in expression: ", base_expr)
            keys = keys[1:]
        if isinstance(value, str):
            unreplaced_count += 1
            continue
        #print var, "->", value 
        expr = expr.replace(var, str(value))
        replaced_count += 1
    try:
        expr = eval(expr)
        expr = float(expr)
    except:
        expr = "%" + expr + "%"

    #print val, "=>", expr
    #print

    return replaced_count, unreplaced_count, expr


# returns number of replaced values
def _parse_config_rec(cfg, top_cfg):

    keys = list()
    if isinstance(cfg, list):
        keys = range(len(cfg))

    elif isinstance(cfg, dict):
        keys = cfg.keys()

    replaced_big_count = 0
    unreplaced_big_count = 0
    for key in keys:
        if isinstance(cfg[key], (float, int, bool)):
            continue
        elif isinstance(cfg[key], str):
            replaced_count, unreplaced_count, res = _parse_val(
                cfg[key], top_cfg)
            replaced_big_count += replaced_count
            unreplaced_big_count += unreplaced_count
            cfg[key] = res
        else:
            replaced_count, unreplaced_count = _parse_config_rec(
                cfg[key], top_cfg)
            replaced_big_count += replaced_count
            unreplaced_big_count += unreplaced_count

    return replaced_big_count, unreplaced_big_count


# overlay 'a' over 'b' and return result
def overlay(a, b):

    if isinstance(a, list):
        assert isinstance(b, list)
        keys = range(len(a))
    elif isinstance(a, dict):
        assert isinstance(b, dict)
        keys = a.keys()

    res = copy.copy(b)

    for key in keys:
        if isinstance(a[key], (float, int, bool)):
            res[key] = a[key]
        elif isinstance(a[key], dict):
            try:
                res[key] = overlay(a[key], b[key])
            except KeyError:
                res[key] = a[key]
        elif isinstance(a[key], list):
            try:
                res[key] = overlay(a[key], b[key])
            except KeyError:
                res[key] = a[key]

    return res



class Distance(float):

    @classmethod
    def from_mm(cls, val):
        return cls( val/1000 )


distance_re = re.compile("^\s*([-\d.]+)(mm|m)\s*$")
def distance_constructor(loader, node):
    value = loader.construct_scalar(node)
    res = re.match(distance_re, value)

    val, unit = res.group(1), res.group(2)

    val = float(val)

    if unit == 'mm':
        return Distance.from_mm(val)
    elif unit == 'm':
        return Distance(val)

yaml.add_constructor(u'!distance', distance_constructor)
yaml.add_implicit_resolver(u'!distance', distance_re)

class Angle(float):

    @classmethod
    def from_degree(cls, val):
        return cls( math.radians(val))

angle_re = re.compile(u"^\s*([-\d.]+)(\xb0*)\s*$")
def angle_constructor(loader, node):
    value = loader.construct_scalar(node)
    res = re.match(angle_re, value)

    val, unit = res.group(1), res.group(2)

    val = float(val)

    if unit == u'\xb0':
        return Angle.from_degree(val)
    elif unit == '':
        return Angle(val)

yaml.add_constructor(u'!angle', angle_constructor)
yaml.add_implicit_resolver(u'!angle', angle_re)

def parse_config(filename):
    config = yaml.load(file(filename, "r"))

    while _parse_config_rec(config, config)[0] > 0:
        pass

    if _parse_config_rec(config, config)[1] > 0:
        raise Exception("Some values in config expressions were not replaced ",
                        config)

    for leg_key in config['legs'].keys():
        config['legs'][leg_key] = overlay(config['legs'][leg_key],
                                          config['default_leg'])

    return config
